"""Main submodule of the library."""
import numpy as np
import quickemcee as qmc


#           Aux stuff, raises etc
class ModelInitException(Exception):
    """Custom exception for Model instance init."""

    def __init__(self, rtext):
        msg = rtext + (' Exception ocurred while trying to instance a tsaopy '
                       'Model object.')
        super().__init__(msg)


#           tsaopy scripts
class BaseModel:
    """
    tsaopy BaseModel class.

    This object condenses all necessary variables to set up the ODE according
    to the parameters provided plus the MCMC configuration to do the fitting.

    """

    def __init__(self, ode_coefs, events, driving_force=None):
        r"""
        Parameters
        ----------
        ode_coefs : dict
            dictionary containing the ode coefficients relevant in the model's
            ODE.

            Note: if the model considers driving force, all three coeffients
            must be included.

            Each key should be one of type of coefficients in the ODE ('a',
            'b', 'f', or 'c') and the value for each key should be a list with
            touples with the index and the prior for each coefficient.

            Indices are given by

            * the order of the term for 'a' and 'b' coefs. Eg: in \(b_2x^2\)
            the index of \(b_2\) coef is 2.
            * for 'f' coefs pass the index of each element in f_coords array
            starting at 1.
            * for 'c' coefs use a touple with two indices for the order of each
            factor. Eg: in \(c_{21}x^2\dot{x}\) index is (2, 1).
        events : list
            list containing all tsaopy Event objects to which the model will be
            fitted. Even if only one Event is used, pass it inside a list.
        driving_force : callable, optional
            A callable to compute the driving force. It must take two arguments
            , first a float for the time variable, and then an array with all
            the parameters other than the time variable. The default is
            None(uses \(F(t)=0\)).

        Examples
        --------
            event1 = tsaopy.events.Event(params1, t1, x1, x1_sigma)
            event2 = tsaopy.events.Event(params2, t2, x2, x2_sigma,
                                         v2, v2_sigma)
            event3 = tsaopy.events.Event(params3, t3, x3, x3_sigma)

            model1_ode_coefs = {'a': [(1, a1_prior), (2, a2_prior)],
                                'b': [(1, b1_prior)],
                                'f': [(1, F_prior), (2, w_prior),
                                      (3, p_prior)],
                                'c': [((2, 1), c21_prior)]}

            model1 = tsaopy.models.Model(ode_coefs=model1_ode_coefs,
                                         events=[event1, event2, event3])
        """
        # run tests
        for i in ode_coefs:
            try:
                ilist = ode_coefs[i]
                # check that every parameter has 2 elements in its touple
                for coef in ilist:
                    if len(coef) < 2:
                        raise ModelInitException('some ODE coef is missing '
                                                 'parameters.')
                    elif len(coef) > 2:
                        raise ModelInitException('some ODE coef has too many '
                                                 'parameters.')
                    # check that for every parameter p(x) > 0
                    else:
                        (_, p), x = coef, np.random.normal(.0, 100.0)
                        if p(x) < .0:
                            raise ModelInitException("some ODE prior returned "
                                                     "a negative value when "
                                                     "called with random "
                                                     "float.")
            # general exception
            except Exception as exception:
                raise ModelInitException('') from exception

        self.odecoefs = []
        self.paramslabels = []
        # a and b 1D vectors
        if 'a' in ode_coefs:
            acoefs = ode_coefs['a']
            acoefs.sort()
            self.odecoefs += acoefs
            self.aind = [c[0] for c in acoefs]
            self.paramslabels += ['a' + str(i) for i in self.aind]
            self.adim = len(self.aind)
            self.alen = max(self.aind)
        else:
            self.aind = []
            self.adim = 0
            self.alen = 0

        if 'b' in ode_coefs:
            bcoefs = ode_coefs['b']
            bcoefs.sort()
            self.odecoefs += bcoefs
            self.bind = [c[0] for c in bcoefs]
            self.paramslabels += ['b' + str(i) for i in self.bind]
            self.bdim = len(self.bind)
            self.blen = max(self.bind)
        else:
            self.bind = []
            self.bdim = 0
            self.blen = 0

        # C matrix
        if 'c' in ode_coefs:
            ccoefs = ode_coefs['c']
            ccoefs.sort()
            self.odecoefs += ccoefs
            self.cind = [c[0] for c in ccoefs]
            self.paramslabels += [r'c_{' + str(i[0]) + str(i[1]) + '}'
                                  for i in self.bind]
            self.cdim = len(self.cind)
            self.cn = max([i[0] for i in self.cind])
            self.cm = max([i[1] for i in self.cind])
        else:
            self.cind = []
            self.cdim = 0
            self.cn, self.cm = 0, 0

        # f ~ driving force
        if 'f' not in ode_coefs and driving_force is None:
            self.using_f = False
            self.fdim = 0
        elif 'f' not in ode_coefs and driving_force is not None:
            raise ModelInitException("driving_force was passed but there is "
                                     "no f key in ode_coefs.")
        elif 'f' in ode_coefs and driving_force is None:
            raise ModelInitException("f key was passed in ode_coefs but "
                                     "driving_force is None.")
        elif 'f' in ode_coefs and driving_force is not None:
            self.using_f = True
            self.df_function = driving_force
            fcoefs = ode_coefs['f']
            fcoefs.sort()
            self.fdim = len(fcoefs)
            self.odecoefs += fcoefs
            self.paramslabels += ['f' + str(i)
                                  for i in range(1, self.fdim + 1)]

        self.odecoefs_ndim = len(self.odecoefs)
        self.odepriors = [c[-1] for c in self.odecoefs]

        # finish param labels & ini guess
        for i, event in enumerate(events):
            self.paramslabels += [str(i+1) + ' - ' + s
                                  for s in ['x0', 'v0']]
            if event.using_ep:
                self.paramslabels.append(str(i+1) + ' - ep')
            if event.using_logfx:
                self.paramslabels.append(str(i+1) + ' - log_fx')
            if event.using_logfv:
                self.paramslabels.append(str(i+1) + ' - log_fv')

        # store events and final data
        self.events = events

        self.ndim = self.odecoefs_ndim
        for e in events:
            self.ndim += e.ndim

    def _ode_arrays(self, abc_coefs):
        alen, blen, cn, cm = self.alen, self.blen, self.cn, self.cm
        aind, bind, cind = self.aind, self.bind, self.cind
        adim, bdim = self.adim, self.bdim
        A, B, C = (np.zeros(alen), np.zeros(blen), np.zeros((cn, cm)))

        for i, ind in enumerate(aind):
            A[ind-1] = abc_coefs[i]
        last_n = adim

        for i, ind in enumerate(bind):
            B[ind-1] = abc_coefs[i + last_n]
        last_n += bdim

        for i, ind in enumerate(cind):
            n, m = ind
            C[n - 1, m - 1] = abc_coefs[i + last_n]

        return A, B, C

    def _df_array(self, event, f_params):
        Flen = 2 * (event.datalen - 1) * event.tsplit + 1
        if self.using_f:
            event_t = np.linspace(event.t_data[0], event.t_data[-1], Flen)
            return self.df_function(event_t, f_params)
        elif not self.using_f:
            return np.zeros(Flen)

    def _log_likelihood(self, coords):
        """Compute log likelihood."""
        abc_dim = self.adim + self.bdim + self.cdim
        f_dim = self.fdim
        abc_coefs, f_params, event_params = (coords[:abc_dim],
                                             coords[abc_dim: abc_dim + f_dim],
                                             coords[abc_dim + f_dim:])

        A, B, C = self._ode_arrays(abc_coefs)

        last_n = 0
        result = .0
        # iterate over all events
        for event in self.events:
            x0v0 = event_params[last_n:last_n + 2]
            last_n += 2
            if event.using_ep:
                ep = event_params[last_n]
                last_n += 1
            else:
                ep = None
            if event.using_logfx:
                logfx = event_params[last_n]
                last_n += 1
            else:
                logfx = None
            if event.using_logfv:
                logfv = event_params[last_n]
                last_n += 1
            else:
                logfv = None

            F = self._df_array(event, f_params)

            result += event._log_likelihood(A, B, C, F, x0v0,
                                            ep, logfx, logfv)

        return result

    def _log_prior(self, coords):
        """Compute log prior."""
        odecoefs_ndim = self.odecoefs_ndim
        odecoords = coords[:odecoefs_ndim]
        result = 1
        for i, p in enumerate(self.odepriors):
            prob = p(odecoords[i])
            if prob <= .0:
                return -np.inf
            result *= prob
        result = np.log(result)

        last_n = odecoefs_ndim
        # iterate over all events
        for event in self.events:
            event_ndim = event.ndim
            event_coords = coords[last_n:last_n + event_ndim]
            last_n += event_ndim
            result += event._log_prior(event_coords)

        return result

    def _log_probability(self, coords):
        """Compute log probability."""
        lp = self._log_prior(coords)
        if not np.isfinite(lp):
            return -np.inf
        return lp + self._log_likelihood(coords)

    def setup_mcmc_model(self):
        """
        Build `quickemcee` model object.

        Returns
        -------
        quickemcee model object
            `quickemcee` model instance that can be used to run MCMC chains.
            See `quickemcee` docs for this class usage.

        Examples
        --------
            tsaopymodel = tsaopy.models.Model(ode_coefs_dict, events_list)
            qmcmodel = tsaopymodel.setup_mcmc_model()
            sampler = qmcmodel.run_chain(100, 50, 100)
        """
        return qmc.core.LPModel(self.ndim, self._log_probability)

    def event_predict(self, i, coords):
        """
        Compute the prediction of the model for the i-eth event.

        This method has the purpose of plotting the prediction of the model for
        one of the events it was fitted to. Used for plotting with the results.

        Parameters
        ----------
        i : int
            An integer 1, 2, 3.. indicating the position of the event to
            simulate in the events list provided to the Model instace. Don't
            start at 0.'
        coords : array
            an array of length ndim(number of ode coefs plus number of params
            for each model, total number of parameters to fit).

        Returns
        -------
        array
            array of the same shape as the x_data attribute in the i-eth event.

        """
        abc_dim = self.adim + self.bdim + self.cdim
        f_dim = self.fdim
        abc_coefs, f_params, event_params = (coords[:abc_dim],
                                             coords[abc_dim: abc_dim + f_dim],
                                             coords[abc_dim + f_dim:])

        A, B, C = self._ode_arrays(abc_coefs)

        last_n = 0
        # iterate over all events
        for j, event in enumerate(self.events):
            if not j == i-1:
                last_n += event.ndim
            elif j == i-1:
                x0v0, ep = event_params[last_n:last_n + 2], None
                last_n += 2
                if event.using_ep:
                    ep = event_params[last_n]
                    last_n += 1
                F = self._df_array(event, f_params)
                return event._predict(A, B, C, F, x0v0, ep)



"""tsaopy Events submodule."""
import numpy as np
from tsaopy._f2pyauxmod import simulation, simulationv


#           Aux stuff, raises etc
class EventInitException(Exception):
    """Custom exception for Event instance init."""

    def __init__(self, rtext):
        msg = rtext + (' Exception ocurred while trying to instance a tsaopy '
                       'Event object.')
        super().__init__(msg)


def _ll(pred, data, sigma):
    return -.5 * np.sum(((pred - data) / sigma) ** 2)


def _ll_logf(pred, data, sigma, logf):
    s2 = sigma ** 2 + pred ** 2 * np.exp(2 * logf)
    return -.5 * np.sum((pred - data) ** 2 / s2 + np.log(s2))


#           tsaopy scripts
class Event:
    """
    tsaopy Event class. Used to later instance tsaopy Model objects.

    The idea behind the Event class is to make it straightforward to fit the
    same equation of motion to different sets of measurements at the same time.

    Suppose we observed the same system several times with different initial
    conditions. Then we will set up an Event instace for each of our experiment
    runs. We will group them all together later when creating a Model class
    instance. With this Model instance we will be fitting all the different
    sets of measurements with the same equation of motion, but different
    initial conditions, equilibrium points, etc.
    """

    def __init__(self, params, t_data, x_data, x_sigma,
                 v_data=None, v_sigma=None):
        """
        Parameters
        ----------
        params : dict
            dictionary containing the parameters relevant in the event. It's
            necessary to always include x0 and v0. Optionally one can use ep
            for the equilibrium point, and log_fx and log_fv for the f term in
            the uncertainty (see docs for more info).

            Each parameter in the dict should be defined with its key ('x0,
            'v0', 'ep', 'log_fx', 'log_fv') and the value should be the prior
            for that parameter.
        t_data : array
            array containing the time values. Values must be evenly spread.
        x_data : array
            array containing the position measurements. Must be of the same
            length as t_data.
        x_sigma : float or array
            uncertainty of the measurements. Can be a float with a unique value
            for all measurements, or an array of the same shape as x_data with
            a unique value for each value of x_data.
        v_data : array, optional
            array containing the velocity measurements. Must be of the same
            length as t_data.
        v_sigma : float or array, optional
            uncertainty of the measurements. Can be a float with a unique value
            for all measurements, or an array of the same shape as v_data with
            a unique value for each value of v_data. Necessary if v_data is
            being used.

        Examples
        --------
            import numpy as np
            import tsaopy
            import quickemcee as qmc

            t = np.linspace(0,10,101)
            x = np.cos(t) + np.random.normal(.0, .3, 101) # data with noise
            x_noise = .3

            x0_prior = qmc.utils.normal_prior(1.0, 10.0)
            v0_prior = qmc.utils.normal_prior(.0, 10.0)

            params_dict = {
                           'x0': x0_prior,
                           'v0': v0_prior
                           }

            event1 = tsaopy.events.Event(params_dict, t, x, x_noise)
        """
        #           TO DO HERE: improve error handling ...
        # make sure that x0 and v0 are in params
        try:
            if 'x0' not in params:
                raise EventInitException('no x0 param when building Event.')
            if 'v0' not in params:
                raise EventInitException('no v0 param when building Event.')
        except Exception as exception:
            raise EventInitException("couldn't verify x0 and v0 keys were incl"
                                     "uded in params dict.") from exception

        # test if priors don't return a positive float when called
        for param in params:
            try:
                x, p = np.random.normal(.0, 100.0), params[param]
                if not np.isfinite(p(x)):
                    raise ValueError("prior for a param return nan or inf when"
                                     " called with a random float.")
                if p(x) < .0:
                    raise ValueError("prior for a param returned a negative va"
                                     "lue when called with a random float.")
            except Exception as exception:
                raise EventInitException("couldn't verify that a random float "
                                         "returns a positive value for some "
                                         "parameter prior.") from exception

        # do check for v data usage
        if v_data is not None:
            if v_sigma is None:
                raise EventInitException("v_data is being used but v_sigma is "
                                         "missing.")
            self.using_v_data = True
        else:
            self.using_v_data = False

        # check x and t have the same lengths, x and v same shape
        try:
            if not len(t_data) == len(x_data):
                raise ValueError('x_data and t_data have different lengths.')
            if self.using_v_data:
                if not x_data.shape == v_data.shape:
                    raise ValueError('x_data and v_data have different shapes.'
                                     )
        except Exception as exception:
            raise EventInitException("couldn't verify that all input arrays "
                                     "have compatible shapes.") from exception

        # check arrays have finite float values
        try:
            if not np.isfinite(x_data).all():
                raise ValueError('x_data has non finite values.')
            if not np.isfinite(t_data).all():
                raise ValueError('t_data has non finite values.')
            if self.using_v_data:
                if not np.isfinite(v_data).all():
                    raise ValueError('v_data has non finite values.')
        except Exception as exception:
            raise EventInitException("couldn't verify that all values in input"
                                     " arrays are finite numbers.") \
                                                            from exception

        #       Define core attributes ~~~~
        self.ndim = len(params)
        self.tsplit = 2
        self.datalen = len(t_data)
        self.dt = (t_data[-1] - t_data[0]) / (self.datalen - 1)

        # attribute params
        self.params = params

        # attibute priors
        self.priors = [params['x0'],
                       params['v0']]

        # do checks for ep
        if 'ep' in params:
            self.using_ep = True
            self.priors.append(params['ep'])
        else:
            self.using_ep = False

        # do checks for fx
        if 'log_fx' in params:
            self.using_logfx = True
            self.priors.append(params['log_fx'])
        else:
            self.using_logfx = False

        # do checks for fv
        if 'log_fv' in params:
            # necessary check
            if not self.using_v_data:
                raise EventInitException("log_fv is in params but there is no "
                                         "v data.")
            self.using_logfv = True
            self.priors.append(params['log_fv'])
        else:
            self.using_logfv = False

        # save observations & sigma data
        self.t_data = t_data
        self.x_data = x_data
        self.x_sigma = x_sigma
        if self.using_v_data:
            self.v_data = v_data
            self.v_sigma = v_sigma

    def _predict(self, A, B, C, F, x0v0, ep):
        """Compute tsaopy prediction using coefs and iniconds."""
        use_ep, use_v = self.using_ep, self.using_v_data
        tsplit = self.tsplit
        dt, datalen = self.dt / tsplit, (self.datalen - 1) * tsplit + 1

        na, nb, nf = len(A), len(B), len(F)
        cn, cm = C.shape

        if cn == 0 or cm == 0:
            cn, cm, C = 1, 1, np.zeros(1)

        if not use_ep and not use_v:
            return simulation(x0v0, A, B, C, F,
                              dt, datalen, na, nb, cn, cm, nf)[::tsplit]
        if use_ep and not use_v:
            return simulation(x0v0, A, B, C, F,
                              dt, datalen, na, nb, cn,
                              cm)[::tsplit] + ep
        if not use_ep and use_v:
            return simulationv(x0v0, A, B, C, F,
                               dt, datalen, na, nb, cn, cm, nf)[::tsplit]
        if use_ep and use_v:
            arr = np.zeros((self.datalen, 2))
            arr[:, 0] += ep
            return simulationv(x0v0, A, B, C, F,
                               dt, datalen, na, nb, cn, cm, nf)[::tsplit] + arr

    def _log_prior(self, event_coords):
        """Compute log prior for event parameters."""
        result = 1
        for i, p in enumerate(self.priors):
            prob = p(event_coords[i])
            if prob <= .0:
                return -np.inf
            result *= prob
        return np.log(result)

    def _log_likelihood(self, A, B, C, F, x0v0,
                        ep, log_fx, log_fv):
        """Compute log likelihood for event parameters and ODE coefs arrays."""
        pred = self._predict(A, B, C, F, x0v0, ep)

        # discard diverging simulations
        if not np.isfinite(pred).all():
            return -np.inf

        use_v, use_fx, use_fv = (self.using_v_data, self.using_logfx,
                                 self.using_logfv)

        # not using v data cases
        if not use_v and not use_fx:
            return _ll(pred, self.x_data, self.x_sigma)
        elif not use_v and use_fx:
            return _ll_logf(pred, self.x_data, self.x_sigma, log_fx)

        predx, predv = pred[:, 0], pred[:, 1]
        # using v data cases
        if use_v and not use_fx and not use_fv:
            return (_ll(predx, self.x_data, self.x_sigma)
                    + _ll(predv, self.v_data, self.v_sigma))
        elif use_v and use_fx and not use_fv:
            return (_ll_logf(predx, self.x_data, self.x_sigma, log_fx)
                    + _ll(predv, self.v_data, self.v_sigma))
        elif use_v and not use_fx and use_fv:
            return (_ll(predx, self.x_data, self.x_sigma)
                    + _ll_logf(predv, self.v_data, self.v_sigma, log_fv))
        elif use_v and use_fx and use_fv:
            return (_ll_logf(predx, self.x_data, self.x_sigma, log_fx)
                    + _ll_logf(predv, self.v_data, self.v_sigma, log_fv))



import os
import sys
import numpy

# cd to file directory
filedir = os.path.dirname(os.path.realpath(__file__))
cwd = os.getcwd()
os.chdir(filedir)

# load fortran source
with open("fortransource.f90", "r") as f90_file:
    fortran_source = f90_file.read()

# compile fortran module
out = numpy.f2py.compile(fortran_source, modulename='_f2pyauxmod',
                         verbose=False, extension='.f90')

if not out == 0:
    sys.exit("Error when building tsaopy submodules, f2py compiler failed.")

os.rename('auxinit.py', '__init__.py')
os.chdir(cwd)

import tsaopy.events
import tsaopy.models



from setuptools import setup

with open("buildreadme.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setup(
      name='tsaopy',
      packages=['tsaopy'],
      version='0.0.5a1',
      description=('Time Series by Anharmonic Oscillators is a Python library '
                   'designed to analize time series by modelling them as '
                   'anharmonic oscillators.'),
      data_files=[('', ['buildreadme.md'])],
      package_data={'tsaopy': ['fortransource.f90']},
      long_description=long_description,
      long_description_content_type="text/markdown",
      url='https://tsaopy.github.io/',
      author='Sofia A. Scozziero',
      author_email='sgscozziero@gmail.com',
      install_requires=['numpy', 'quickemcee'],
      classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: POSIX :: Linux", ]
)

